import asyncio
import copy
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Dict, List, Optional, Set

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2 import transformations as tr
from arcor2.cached import CachedProject, CachedScene, UpdateableCachedProject
from arcor2.data import common
from arcor2.data.events import Event, PackageState
from arcor2.exceptions import Arcor2Exception
from arcor2.logic import LogicContainer, check_for_loops
from arcor2.object_types.abstract import Generic, Robot
from arcor2.parameter_plugins.base import ParameterPluginException
from arcor2.parameter_plugins.pose import PosePlugin
from arcor2.parameter_plugins.utils import plugin_from_type_name
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import project
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.helpers import (
    ctx_read_lock,
    ctx_write_lock,
    ensure_write_locked,
    get_unlocked_objects,
    make_name_unique,
    unique_name,
)
from arcor2_arserver.objects_actions import get_types_dict
from arcor2_arserver.project import (
    check_action_params,
    check_flows,
    close_project,
    find_object_action,
    notify_project_opened,
    open_project,
    project_names,
    project_problems,
)
from arcor2_arserver.robot import check_eef_arm, get_end_effector_pose, get_pose_and_joints, get_robot_joints
from arcor2_arserver.scene import can_modify_scene, ensure_scene_started, get_instance, get_robot_instance, open_scene
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc


@asynccontextmanager
async def managed_project(project_id: str, make_copy: bool = False) -> AsyncGenerator[UpdateableCachedProject, None]:

    save_back = False

    if glob.LOCK.project and glob.LOCK.project.id == project_id:
        if make_copy:
            project = copy.deepcopy(glob.LOCK.project)
            save_back = True
        else:
            project = glob.LOCK.project
    else:
        save_back = True
        project = UpdateableCachedProject(await storage.get_project(project_id))

    if make_copy:
        project.id = common.Project.uid()

    try:
        yield project
    finally:
        if save_back:
            asyncio.ensure_future(storage.update_project(project))


async def cancel_action_cb(req: srpc.p.CancelAction.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    if not glob.RUNNING_ACTION:
        raise Arcor2Exception("No action is running.")

    action = proj.action(glob.RUNNING_ACTION)
    obj_id, action_name = action.parse_type()
    obj = get_instance(obj_id, Generic)

    if not find_object_action(scene, action).meta.cancellable:
        raise Arcor2Exception("Action is not cancellable.")

    try:
        action_method = getattr(obj, action_name)
        cancel_method = getattr(obj, obj.CANCEL_MAPPING[action_method.__name__])
    except AttributeError as e:
        raise Arcor2Exception("Internal error.") from e

    cancel_params: Dict[str, Any] = {}

    # TODO fix or remove this?
    """
    cancel_sig = inspect.signature(cancel_method)

    assert glob.RUNNING_ACTION_PARAMS is not None

    for param_name in cancel_sig.parameters.keys():
        try:
            cancel_params[param_name] = glob.RUNNING_ACTION_PARAMS[param_name]
        except KeyError as e:
            raise Arcor2Exception("Cancel method parameters should be subset of action parameters.") from e
    """

    await hlp.run_in_executor(cancel_method, *cancel_params.values())

    asyncio.ensure_future(notif.broadcast_event(sevts.a.ActionCancelled()))
    glob.RUNNING_ACTION = None
    glob.RUNNING_ACTION_PARAMS = None


async def execute_action_cb(req: srpc.p.ExecuteAction.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    # TODO rather lock the project and release the lock once execution is finished?
    async with ctx_write_lock(glob.LOCK.SpecialValues.RUNNING_ACTION, glob.USERS.user_name(ui)):

        ensure_scene_started()

        if glob.RUNNING_ACTION:
            raise Arcor2Exception(
                f"Action {glob.RUNNING_ACTION} is being executed. " f"Only one action can be executed at a time."
            )

        action = proj.action(req.args.action_id)

        obj_id, action_name = action.parse_type()

        params: List[Any] = []

        for param in action.parameters:

            if param.type == common.ActionParameter.TypeEnum.LINK:

                parsed_link = param.parse_link()
                try:
                    results = glob.PREV_RESULTS[parsed_link.action_id]
                except KeyError:
                    prev_action = proj.action(parsed_link.action_id)
                    raise Arcor2Exception(f"Action '{prev_action.name}' has to be executed first.")

                # an action result could be a tuple or a single value
                if isinstance(results, tuple):
                    params.append(results[parsed_link.output_index])
                else:
                    assert parsed_link.output_index == 0
                    params.append(results)

            elif param.type == common.ActionParameter.TypeEnum.PROJECT_PARAMETER:
                pparam = proj.parameter(param.str_from_value())
                # TODO use plugin to get the value
                from arcor2 import json

                params.append(json.loads(pparam.value))
            else:

                try:
                    params.append(
                        plugin_from_type_name(param.type).parameter_execution_value(
                            get_types_dict(), scene, proj, action.id, param.name
                        )
                    )
                except ParameterPluginException as e:
                    glob.logger.error(e)
                    raise Arcor2Exception(f"Failed to get value for parameter {param.name}.")

        obj = get_instance(obj_id, Generic)

        if isinstance(obj, Robot):
            if obj.move_in_progress:
                raise Arcor2Exception("Can't execute actions while the robot moves.")

        if not hasattr(obj, action_name):
            raise Arcor2Exception("Internal error: object does not have the requested method.")

        if req.dry_run:
            return

        glob.RUNNING_ACTION = action.id
        glob.RUNNING_ACTION_PARAMS = params

        glob.logger.debug(f"Running action {action.name} ({type(obj)}/{action_name}), params: {params}.")

        # schedule execution and return success
        asyncio.ensure_future(project.execute_action(getattr(obj, action_name), params))
        return None


async def project_info(
    project_id: str, scenes_lock: asyncio.Lock, scenes: Dict[str, CachedScene]
) -> srpc.p.ListProjects.Response.Data:

    project = await storage.get_project(project_id)

    assert project.created
    assert project.modified

    pd = srpc.p.ListProjects.Response.Data(
        project.name,
        project.scene_id,
        project.description,
        project.has_logic,
        project.created,
        project.modified,
        id=project.id,
    )

    try:
        async with scenes_lock:
            if project.scene_id not in scenes:
                scenes[project.scene_id] = await storage.get_scene(project.scene_id)
    except storage.ProjectServiceException:
        pd.problems.append("Scene does not exist.")
        return pd

    pd.problems = project_problems(scenes[project.scene_id], project)
    pd.valid = not pd.problems

    if not pd.valid:
        return pd

    try:
        # TODO call build service!!!
        pd.executable = True
    except Arcor2Exception as e:
        pd.problems.append(str(e))

    return pd


async def list_projects_cb(req: srpc.p.ListProjects.Request, ui: WsClient) -> srpc.p.ListProjects.Response:

    scenes_lock = asyncio.Lock()
    scenes: Dict[str, CachedScene] = {}

    resp = srpc.p.ListProjects.Response()
    tasks = [project_info(proj_id, scenes_lock, scenes) for proj_id in (await storage.get_project_ids())]

    resp.data = []
    for res in await asyncio.gather(*tasks, return_exceptions=True):

        if isinstance(res, Arcor2Exception):
            glob.logger.error(str(res))
        elif isinstance(res, Exception):
            raise res  # zero toleration for other exceptions
        else:
            resp.data.append(res)

    return resp


async def add_ap_using_robot_cb(req: srpc.p.AddApUsingRobot.Request, ui: WsClient) -> None:
    async def notify(ap: common.BareActionPoint, ori: common.NamedOrientation, joi: common.ProjectRobotJoints) -> None:

        ap_evt = sevts.p.ActionPointChanged(ap)
        ap_evt.change_type = Event.Type.ADD
        await notif.broadcast_event(ap_evt)

        ori_evt = sevts.p.OrientationChanged(ori)
        ori_evt.change_type = Event.Type.ADD
        ori_evt.parent_id = ap.id

        joi_evt = sevts.p.JointsChanged(joi)
        joi_evt.change_type = Event.Type.ADD
        joi_evt.parent_id = ap.id

        await asyncio.gather(notif.broadcast_event(ori_evt), notif.broadcast_event(joi_evt))

    hlp.is_valid_identifier(req.args.name)
    proj = glob.LOCK.project_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):

        ensure_scene_started()

        unique_name(req.args.name, proj.action_points_names)

        robot_inst = get_robot_instance(req.args.robot_id)
        await check_eef_arm(robot_inst, req.args.arm_id, req.args.end_effector_id)

        if req.dry_run:
            return None

        pose, joints = await get_pose_and_joints(robot_inst, req.args.end_effector_id, req.args.arm_id)

        ap = proj.upsert_action_point(common.ActionPoint.uid(), req.args.name, pose.position)
        ori = common.NamedOrientation("default", pose.orientation)
        proj.upsert_orientation(ap.id, ori)
        joi = common.ProjectRobotJoints("default", req.args.robot_id, joints, True, req.args.arm_id)
        proj.upsert_joints(ap.id, joi)

        asyncio.ensure_future(notify(ap, ori, joi))
        return None


async def add_action_point_joints_using_robot_cb(
    req: srpc.p.AddActionPointJointsUsingRobot.Request, ui: WsClient
) -> None:

    hlp.is_valid_identifier(req.args.name)
    proj = glob.LOCK.project_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):

        ensure_scene_started()

        robot_inst = get_robot_instance(req.args.robot_id)
        await check_eef_arm(robot_inst, req.args.arm_id)

        ap = proj.bare_action_point(req.args.action_point_id)

        unique_name(req.args.name, proj.ap_joint_names(ap.id))

        new_joints = await get_robot_joints(robot_inst, req.args.arm_id)

        await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

        if req.dry_run:
            return None

        prj = common.ProjectRobotJoints(req.args.name, req.args.robot_id, new_joints, True, req.args.arm_id)
        proj.upsert_joints(ap.id, prj)

        evt = sevts.p.JointsChanged(prj)
        evt.change_type = Event.Type.ADD
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_point_joints_using_robot_cb(
    req: srpc.p.UpdateActionPointJointsUsingRobot.Request, ui: WsClient
) -> None:

    proj = glob.LOCK.project_or_exception()

    ensure_scene_started()

    ap, robot_joints = proj.ap_and_joints(req.args.joints_id)

    user_name = glob.USERS.user_name(ui)

    async with ctx_read_lock(robot_joints.robot_id, user_name):
        await ensure_write_locked(ap.id, user_name)

        robot_joints.joints = await get_robot_joints(get_robot_instance(robot_joints.robot_id), robot_joints.arm_id)
        robot_joints.is_valid = True

        proj.update_modified()

        evt = sevts.p.JointsChanged(robot_joints)
        evt.change_type = Event.Type.UPDATE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_point_joints_cb(req: srpc.p.UpdateActionPointJoints.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    ap, robot_joints = proj.ap_and_joints(req.args.joints_id)

    if {joint.name for joint in req.args.joints} != {joint.name for joint in robot_joints.joints}:
        raise Arcor2Exception("Joint names does not match the robot.")

    await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

    # TODO maybe joints values should be normalized? To <0, 2pi> or to <-pi, pi>?
    robot_joints.joints = req.args.joints
    robot_joints.is_valid = True
    proj.update_modified()

    evt = sevts.p.JointsChanged(robot_joints)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def remove_action_point_joints_cb(req: srpc.p.RemoveActionPointJoints.Request, ui: WsClient) -> None:
    """Removes joints from action point.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()

    for act in proj.actions:
        for param in act.parameters:
            if plugin_from_type_name(param.type).uses_robot_joints(proj, act.id, param.name, req.args.joints_id):
                raise Arcor2Exception(f"Joints used in action {act.name} (parameter {param.name}).")

    ap, _ = proj.ap_and_joints(req.args.joints_id)
    await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

    joints_to_be_removed = proj.remove_joints(req.args.joints_id)

    proj.update_modified()

    evt = sevts.p.JointsChanged(joints_to_be_removed)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def rename_action_point_cb(req: srpc.p.RenameActionPoint.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    ap = proj.bare_action_point(req.args.action_point_id)

    if req.args.new_name == ap.name:
        raise Arcor2Exception("Name unchanged")

    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, proj.action_points_names)

    user_name = glob.USERS.user_name(ui)
    await ensure_write_locked(req.args.action_point_id, user_name)

    if req.dry_run:
        return None

    ap.name = req.args.new_name

    proj.update_modified()

    asyncio.create_task(glob.LOCK.write_unlock(req.args.action_point_id, user_name, True))

    evt = sevts.p.ActionPointChanged(ap)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


def detect_ap_loop(proj: CachedProject, ap: common.BareActionPoint, new_parent_id: str) -> None:

    visited_ids: Set[str] = set()
    ap = copy.deepcopy(ap)
    ap.parent = new_parent_id
    while True:

        if ap.id in visited_ids:
            raise Arcor2Exception("Loop detected!")

        visited_ids.add(ap.id)

        if ap.parent is None:
            break  # type: ignore  # this is certainly reachable

        try:
            ap = proj.bare_action_point(ap.parent)
        except Arcor2Exception:
            break


async def update_action_point_parent_cb(req: srpc.p.UpdateActionPointParent.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)
    ap = proj.bare_action_point(req.args.action_point_id)

    # some super basic checks (not involving other objects) are performed before locking anything
    if req.args.new_parent_id == ap.parent:
        return None

    if req.args.new_parent_id == ap.id:
        raise Arcor2Exception("AP can't be its own parent.")

    to_lock = await get_unlocked_objects(ap.parent, user_name) if ap.parent else set()
    if req.args.new_parent_id:
        to_lock.add(req.args.new_parent_id)

    async with ctx_write_lock(to_lock, user_name):

        check_ap_parent(scene, proj, req.args.new_parent_id)
        detect_ap_loop(proj, ap, req.args.new_parent_id)

        await ensure_write_locked(ap.id, user_name)

        if req.dry_run:
            return

        # Save root and parent of current AP and apply it to structures after successful update
        current_root = await glob.LOCK.get_root_id(ap.id)
        old_parent = ap.parent

        if not ap.parent and req.args.new_parent_id:
            # AP position and all orientations will become relative to the parent
            tr.make_global_ap_relative(scene, proj, ap, req.args.new_parent_id)

        elif ap.parent and not req.args.new_parent_id:
            # AP position and all orientations will become absolute
            tr.make_relative_ap_global(scene, proj, ap)
        else:

            assert ap.parent
            # AP position and all orientations will become relative to another parent
            tr.make_relative_ap_global(scene, proj, ap)
            tr.make_global_ap_relative(scene, proj, ap, req.args.new_parent_id)

        proj.update_child(ap.id, old_parent, req.args.new_parent_id)
        await glob.LOCK.update_write_lock(ap.id, current_root, user_name)

        assert (req.args.new_parent_id and ap.parent == req.args.new_parent_id) or (
            req.args.new_parent_id == "" and ap.parent is None
        )
        proj.update_modified()

        """
        Can't send orientation changes and then ActionPointChanged/UPDATE_BASE (or vice versa)
        because UI would display orientations wrongly (for a short moment).
        """
        # 'ap' is BareActionPoint, that does not contain orientations
        evt = sevts.p.ActionPointChanged(proj.action_point(req.args.action_point_id))
        evt.change_type = Event.Type.UPDATE
        asyncio.ensure_future(notif.broadcast_event(evt))

    asyncio.create_task(glob.LOCK.write_unlock(ap.id, user_name, True))

    return None


async def update_ap_position(
    proj: UpdateableCachedProject, ap: common.BareActionPoint, position: common.Position
) -> None:
    """Updates position of an AP and sends notification about joints that
    become invalid because of it.

    :param ap:
    :param position:
    :return:
    """

    valid_joints = [joints for joints in proj.ap_joints(ap.id) if joints.is_valid]
    proj.update_ap_position(ap.id, position)

    for joints in valid_joints:  # those are now invalid, so let's notify UI about the change

        assert not joints.is_valid

        evt = sevts.p.JointsChanged(joints)
        evt.change_type = Event.Type.UPDATE
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))

    ap_evt = sevts.p.ActionPointChanged(ap)
    ap_evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(ap_evt))


async def update_action_point_position_cb(req: srpc.p.UpdateActionPointPosition.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    ap = proj.bare_action_point(req.args.action_point_id)
    user_name = glob.USERS.user_name(ui)

    if req.dry_run:
        await glob.LOCK.check_lock_tree(req.args.action_point_id, user_name)
    else:
        await ensure_write_locked(req.args.action_point_id, user_name, True)

    if req.dry_run:
        return

    await update_ap_position(proj, ap, req.args.new_position)


async def update_action_point_using_robot_cb(req: srpc.p.UpdateActionPointUsingRobot.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    ensure_scene_started()
    user_name = glob.USERS.user_name(ui)

    ap = proj.bare_action_point(req.args.action_point_id)
    await ensure_write_locked(ap.id, user_name)

    async with ctx_read_lock(req.args.robot.robot_id, user_name):
        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)
        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        if ap.parent:
            new_pose = tr.make_pose_rel_to_parent(scene, proj, new_pose, ap.parent)

        await update_ap_position(proj, proj.bare_action_point(req.args.action_point_id), new_pose.position)
        return None


async def add_action_point_orientation_cb(req: srpc.p.AddActionPointOrientation.Request, ui: WsClient) -> None:
    """Adds orientation and joints to the action point.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()

    ap = proj.bare_action_point(req.args.action_point_id)
    hlp.is_valid_identifier(req.args.name)
    unique_name(req.args.name, proj.ap_orientation_names(ap.id))

    await ensure_write_locked(req.args.action_point_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    orientation = common.NamedOrientation(req.args.name, req.args.orientation)
    proj.upsert_orientation(ap.id, orientation)

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.ADD
    evt.parent_id = ap.id
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def update_action_point_orientation_cb(req: srpc.p.UpdateActionPointOrientation.Request, ui: WsClient) -> None:
    """Updates orientation of the action point.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()
    orientation = proj.orientation(req.args.orientation_id)
    await ensure_write_locked(orientation.id, glob.USERS.user_name(ui))

    orientation.orientation = req.args.orientation

    proj.update_modified()

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def add_action_point_orientation_using_robot_cb(
    req: srpc.p.AddActionPointOrientationUsingRobot.Request, ui: WsClient
) -> None:
    """Adds orientation and joints to the action point.

    :param req:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_read_lock(req.args.robot.robot_id, user_name):

        ensure_scene_started()

        ap = proj.bare_action_point(req.args.action_point_id)
        hlp.is_valid_identifier(req.args.name)
        unique_name(req.args.name, proj.ap_orientation_names(ap.id))
        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)

        await ensure_write_locked(req.args.action_point_id, user_name)

        if req.dry_run:
            return None

        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        if ap.parent:
            new_pose = tr.make_pose_rel_to_parent(scene, proj, new_pose, ap.parent)

        orientation = common.NamedOrientation(req.args.name, new_pose.orientation)
        proj.upsert_orientation(ap.id, orientation)

        evt = sevts.p.OrientationChanged(orientation)
        evt.change_type = Event.Type.ADD
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_point_orientation_using_robot_cb(
    req: srpc.p.UpdateActionPointOrientationUsingRobot.Request, ui: WsClient
) -> None:
    """Updates orientation and joint of the action point.

    :param req:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_read_lock(req.args.robot.robot_id, user_name):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)

        ap, ori = proj.bare_ap_and_orientation(req.args.orientation_id)

        await ensure_write_locked(ori.id, user_name)

        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        if ap.parent:
            new_pose = tr.make_pose_rel_to_parent(scene, proj, new_pose, ap.parent)

        ori.orientation = new_pose.orientation

        proj.update_modified()

        evt = sevts.p.OrientationChanged(ori)
        evt.change_type = Event.Type.UPDATE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def remove_action_point_orientation_cb(req: srpc.p.RemoveActionPointOrientation.Request, ui: WsClient) -> None:
    """Removes orientation.

    :param req:
    :return:
    """

    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)
    ap, orientation = proj.bare_ap_and_orientation(req.args.orientation_id)

    to_lock = await get_unlocked_objects(orientation.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):
        for act in proj.actions:
            for param in act.parameters:
                if plugin_from_type_name(param.type).uses_orientation(
                    proj, act.id, param.name, req.args.orientation_id
                ):
                    raise Arcor2Exception(f"Orientation used in action {act.name} (parameter {param.name}).")

        if req.dry_run:
            return None

        await glob.LOCK.write_unlock(orientation.id, user_name)

        proj.remove_orientation(req.args.orientation_id)

        evt = sevts.p.OrientationChanged(orientation)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def open_project_cb(req: srpc.p.OpenProject.Request, ui: WsClient) -> None:
    """Opens a project. This can be done even when a scene or another project
    is opened.

    :param req:
    :param ui:
    :return:
    """

    async with glob.LOCK.get_lock():

        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't open project while package runs.")

        # TODO validate using project_problems?
        await open_project(req.args.id)

        assert glob.LOCK.scene
        assert glob.LOCK.project

        asyncio.ensure_future(
            notify_project_opened(
                sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.LOCK.scene.scene, glob.LOCK.project.project))
            )
        )

        return None


async def save_project_cb(req: srpc.p.SaveProject.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock(req.dry_run):

        proj = glob.LOCK.project_or_exception()

        if proj.modified and not proj.has_changes:
            raise Arcor2Exception("No changes to save.")

        if req.dry_run:
            return None

        # TODO temporary code to help debugging long lasting update of project
        import time

        start = time.monotonic()
        proj.modified = await storage.update_project(proj)
        glob.logger.info(f"Updating the project took {time.monotonic()-start:.3f}s.")

    asyncio.ensure_future(notif.broadcast_event(sevts.p.ProjectSaved()))
    return None


async def new_project_cb(req: srpc.p.NewProject.Request, ui: WsClient) -> None:

    if glob.LOCK.project:
        raise Arcor2Exception("Project has to be closed first.")

    async with ctx_write_lock(glob.LOCK.SpecialValues.PROJECT_NAME, glob.USERS.user_name(ui), dry_run=req.dry_run):
        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't create project while package runs.")

        unique_name(req.args.name, (await project_names()))

        if await glob.LOCK.get_write_locks_count() > 1:  # project lock also counts
            raise Arcor2Exception("Project has locked objects")

        if req.dry_run:
            return None

        if glob.LOCK.scene:
            if glob.LOCK.scene.id != req.args.scene_id:
                raise Arcor2Exception("Another scene is opened.")

            if glob.LOCK.scene.has_changes:
                glob.LOCK.scene.modified = await storage.update_scene(glob.LOCK.scene)
        else:

            if req.args.scene_id not in (await storage.get_scene_ids()):
                raise Arcor2Exception("Unknown scene id.")

            await open_scene(req.args.scene_id)

        glob.PREV_RESULTS.clear()
        glob.LOCK.project = UpdateableCachedProject(
            common.Project(
                req.args.name, req.args.scene_id, description=req.args.description, has_logic=req.args.has_logic
            )
        )

        assert glob.LOCK.scene

        # add commonly used project parameters
        from arcor2 import json  # TODO use parameter plugin

        glob.LOCK.project.upsert_parameter(
            common.ProjectParameter("scene_id", "string", json.dumps(glob.LOCK.scene.id))
        )
        glob.LOCK.project.upsert_parameter(
            common.ProjectParameter("project_id", "string", json.dumps(glob.LOCK.project.id))
        )

        asyncio.ensure_future(
            notify_project_opened(
                sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.LOCK.scene.scene, glob.LOCK.project.project))
            )
        )
        return None


async def close_project_cb(req: srpc.p.CloseProject.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock(req.dry_run):

        proj = glob.LOCK.project_or_exception()
        can_modify_scene()

        if not req.args.force and proj.has_changes:
            raise Arcor2Exception("Project has unsaved changes.")

        if req.dry_run:
            return None

        await close_project()
        return None


def check_ap_parent(scene: CachedScene, proj: CachedProject, parent: Optional[str]) -> None:

    if not parent:
        return

    if parent in scene.object_ids:
        if scene.object(parent).pose is None:
            raise Arcor2Exception("AP can't have object without pose as parent.")
    elif parent not in proj.action_points_ids:
        raise Arcor2Exception("AP has invalid parent ID (not an object or another AP).")


async def add_action_point_cb(req: srpc.p.AddActionPoint.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    async with ctx_write_lock(glob.LOCK.SpecialValues.PROJECT_NAME, glob.USERS.user_name(ui)):

        hlp.is_valid_identifier(req.args.name)
        unique_name(req.args.name, proj.action_points_names)
        check_ap_parent(scene, proj, req.args.parent)

        if req.dry_run:
            return None

        ap = proj.upsert_action_point(common.ActionPoint.uid(), req.args.name, req.args.position, req.args.parent)

        evt = sevts.p.ActionPointChanged(ap)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def remove_action_point_cb(req: srpc.p.RemoveActionPoint.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        ap = proj.bare_action_point(req.args.id)

        for proj_ap in proj.action_points_with_parent:
            if proj_ap.parent == ap.id:
                raise Arcor2Exception(f"Can't remove parent of '{proj_ap.name}' AP.")

        ap_action_ids = proj.ap_action_ids(ap.id)

        # check if AP's actions aren't involved in logic
        # TODO 'force' param to remove logical connections?
        for logic in proj.logic:
            if (
                logic.start in ap_action_ids
                or logic.end in ap_action_ids
                or (logic.condition and logic.condition.parse_what().action_id in ap_action_ids)
            ):
                raise Arcor2Exception("Remove logic connections first.")

        for act in proj.actions:

            if act.id in ap_action_ids:
                continue

            for param in act.parameters:
                if param.type == common.ActionParameter.TypeEnum.LINK:
                    parsed_link = param.parse_link()
                    linking_action = proj.action(parsed_link.action_id)
                    if parsed_link.action_id in ap_action_ids:
                        raise Arcor2Exception(f"Result of '{act.name}' is linked from '{linking_action.name}'.")

                if not param.is_value():
                    continue

                for joints in proj.ap_joints(ap.id):
                    if plugin_from_type_name(param.type).uses_robot_joints(proj, act.id, param.name, joints.id):
                        raise Arcor2Exception(
                            f"Joints {joints.name} used in action {act.name} (parameter {param.name})."
                        )

                for ori in proj.ap_orientations(ap.id):
                    if plugin_from_type_name(param.type).uses_orientation(proj, act.id, param.name, ori.id):
                        raise Arcor2Exception(
                            f"Orientation {ori.name} used in action {act.name} (parameter {param.name})."
                        )

                # TODO some hypothetical parameter type could use just bare ActionPoint (its position)

        if not await glob.LOCK.check_remove(ap.id, user_name):
            raise Arcor2Exception("Children locked")

        if req.dry_run:
            return None

        await glob.LOCK.write_unlock(req.args.id, user_name)

        proj.remove_action_point(req.args.id)

        evt = sevts.p.ActionPointChanged(ap)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def copy_action_point_cb(req: srpc.p.CopyActionPoint.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    async def copy_action_point(
        orig_ap: common.BareActionPoint,
        new_parent_id: Optional[str] = None,
        position: Optional[common.Position] = None,
    ) -> None:

        ap = proj.upsert_action_point(
            common.ActionPoint.uid(),
            make_name_unique(f"{orig_ap.name}_copy", proj.action_points_names),
            orig_ap.position if position is None else position,
            orig_ap.parent if new_parent_id is None else new_parent_id,
        )

        ap_added_evt = sevts.p.ActionPointChanged(ap)
        ap_added_evt.change_type = Event.Type.ADD
        await notif.broadcast_event(ap_added_evt)

        for ori in proj.ap_orientations(orig_ap.id):
            new_ori = ori.copy()
            old_ori_to_new_ori[ori.id] = new_ori.id
            proj.upsert_orientation(ap.id, new_ori)

            ori_added_evt = sevts.p.OrientationChanged(new_ori)
            ori_added_evt.change_type = Event.Type.ADD
            ori_added_evt.parent_id = ap.id
            await notif.broadcast_event(ori_added_evt)

        for joints in proj.ap_joints(orig_ap.id):
            new_joints = joints.copy()
            proj.upsert_joints(ap.id, new_joints)

            joints_added_evt = sevts.p.JointsChanged(new_joints)
            joints_added_evt.change_type = Event.Type.ADD
            joints_added_evt.parent_id = ap.id
            await notif.broadcast_event(joints_added_evt)

        action_names = proj.action_names  # action name has to be globally unique
        for act in proj.ap_actions(orig_ap.id):
            new_act = act.copy()
            new_act.name = make_name_unique(f"{act.name}_copy", action_names)
            proj.upsert_action(ap.id, new_act)

            for param in new_act.parameters:

                if param.type != PosePlugin.type_name():
                    continue

                old_ori_id = PosePlugin.orientation_id(proj, new_act.id, param.name)

                # TODO this is hacky - plugins are missing methods to set/update parameters
                from arcor2 import json

                # TODO this won't work if action on AP is using orientation from AP's descendant
                #  ...which is not in the mapping yet
                try:
                    param.value = json.dumps(old_ori_to_new_ori[old_ori_id])
                except KeyError:
                    glob.logger.error(f"Failed to find a new orientation ID for {old_ori_id}.")

            action_added_evt = sevts.p.ActionChanged(new_act)
            action_added_evt.change_type = Event.Type.ADD
            action_added_evt.parent_id = ap.id
            await notif.broadcast_event(action_added_evt)

        for child_ap in proj.action_points_with_parent:
            if child_ap.parent == orig_ap.id:
                await copy_action_point(child_ap, ap.id)

    async with ctx_read_lock(req.args.id, glob.USERS.user_name(ui)):
        original_ap = proj.bare_action_point(req.args.id)

        if req.dry_run:
            return

        old_ori_to_new_ori: Dict[str, str] = {}
        asyncio.ensure_future(copy_action_point(original_ap, position=req.args.position))


async def add_action_cb(req: srpc.p.AddAction.Request, ui: WsClient) -> None:
    """Adds new action to project.

    Used also when duplicating action.
    """

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    # When duplicating action AP cannot be removed, no need to lock
    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.action_point_id, user_name)
    async with ctx_write_lock(to_lock, user_name):
        ap = proj.bare_action_point(req.args.action_point_id)

        unique_name(req.args.name, proj.action_names)

        new_action = common.Action(req.args.name, req.args.type, parameters=req.args.parameters, flows=req.args.flows)

        action_meta = find_object_action(scene, new_action)

        updated_project = copy.deepcopy(proj)
        updated_project.upsert_action(req.args.action_point_id, new_action)

        check_flows(updated_project, new_action, action_meta)
        check_action_params(scene, updated_project, new_action, action_meta)

        if req.dry_run:
            return None

        proj.upsert_action(ap.id, new_action)

        evt = sevts.p.ActionChanged(new_action)
        evt.change_type = Event.Type.ADD
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_action_cb(req: srpc.p.UpdateAction.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    updated_project = copy.deepcopy(proj)

    updated_action = updated_project.action(req.args.action_id)

    if req.args.parameters is not None:
        updated_action.parameters = req.args.parameters
    if req.args.flows is not None:
        updated_action.flows = req.args.flows

    updated_action_meta = find_object_action(scene, updated_action)

    check_flows(updated_project, updated_action, updated_action_meta)
    check_action_params(scene, updated_project, updated_action, updated_action_meta)

    await ensure_write_locked(req.args.action_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    orig_action = proj.action(req.args.action_id)
    orig_action.parameters = updated_action.parameters
    proj.update_modified()

    evt = sevts.p.ActionChanged(updated_action)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


def check_action_usage(proj: CachedProject, action: common.Action) -> None:

    # check parameters
    for act in proj.actions:
        for param in act.parameters:
            if param.type == common.ActionParameter.TypeEnum.LINK:
                link = param.parse_link()
                if action.id == link.action_id:
                    raise Arcor2Exception(f"Action output used as parameter of {act.name}/{param.name}.")

    # check logic
    for log in proj.logic:

        if log.start == action.id or log.end == action.id:
            raise Arcor2Exception("Action used in logic.")

        if log.condition:
            action_id, _, _ = log.condition.what.split("/")

            if action_id == action.id:
                raise Arcor2Exception("Action used in condition.")


async def remove_action_cb(req: srpc.p.RemoveAction.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        ap, action = proj.action_point_and_action(req.args.id)
        check_action_usage(proj, action)

        if req.dry_run:
            return None

        await glob.LOCK.write_unlock(req.args.id, user_name)

        proj.remove_action(req.args.id)
        glob.remove_prev_result(action.id)

        evt = sevts.p.ActionChanged(action.bare)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


def check_logic_item(scene: CachedScene, parent: LogicContainer, logic_item: common.LogicItem) -> None:
    """Checks if newly added/updated ProjectLogicItem is ok.

    :param parent:
    :param logic_item:
    :return:
    """

    action_ids = parent.action_ids()

    if logic_item.start == common.LogicItem.START and logic_item.end == common.LogicItem.END:
        raise Arcor2Exception("This does not make sense.")

    if logic_item.start != common.LogicItem.START:

        start_action_id, start_flow = logic_item.parse_start()

        if start_action_id == logic_item.end:
            raise Arcor2Exception("Start and end can't be the same.")

        if start_action_id not in action_ids:
            raise Arcor2Exception("Logic item has unknown start.")

        if start_flow != "default":
            raise Arcor2Exception("Only flow 'default' is supported so far.'")

    if logic_item.end != common.LogicItem.END:

        if logic_item.end not in action_ids:
            raise Arcor2Exception("Logic item has unknown end.")

    if logic_item.condition is not None:

        what = logic_item.condition.parse_what()
        action = parent.action(what.action_id)  # action that produced the result which we depend on here
        flow = action.flow(what.flow_name)
        try:
            flow.outputs[what.output_index]
        except IndexError:
            raise Arcor2Exception(f"Flow {flow.type} does not have output with index {what.output_index}.")

        action_meta = find_object_action(scene, action)

        try:
            return_type = action_meta.returns[what.output_index]
        except IndexError:
            raise Arcor2Exception(f"Invalid output index {what.output_index} for action {action_meta.name}.")

        return_type_plugin = plugin_from_type_name(return_type)

        if not return_type_plugin.COUNTABLE:
            raise Arcor2Exception(f"Output of type {return_type} can't be branched.")

        # TODO for now there is only support for bool
        if return_type_plugin.type() != bool:
            raise Arcor2Exception("Unsupported condition type.")

        # check that condition value is ok, actual value is not interesting
        # TODO perform this check using plugin
        from arcor2 import json

        if not isinstance(json.loads(logic_item.condition.value), bool):
            raise Arcor2Exception("Invalid condition value.")

    for existing_item in parent.logic:

        if existing_item.id == logic_item.id:  # item is updated
            continue

        if logic_item.start == logic_item.START and existing_item.start == logic_item.START:
            raise Arcor2Exception("START already defined.")

        if logic_item.start == existing_item.start:

            if None in (logic_item.condition, existing_item.condition):
                raise Arcor2Exception("Two junctions has the same start action without condition.")

            # when there are more logical connections from A to B, their condition values must be different
            if logic_item.condition == existing_item.condition:
                raise Arcor2Exception("Two junctions with the same start should have different conditions.")

        if logic_item.end == existing_item.end:
            if logic_item.start == existing_item.start:
                raise Arcor2Exception("Junctions can't have the same start and end.")


async def add_logic_item_cb(req: srpc.p.AddLogicItem.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()
    user_name = glob.USERS.user_name(ui)

    to_lock = await get_unlocked_objects([req.args.start, req.args.end], user_name)
    async with ctx_write_lock(to_lock, user_name):
        logic_item = common.LogicItem(req.args.start, req.args.end, req.args.condition)
        check_logic_item(scene, proj, logic_item)

        if logic_item.start != logic_item.START:
            updated_project = copy.deepcopy(proj)
            updated_project.upsert_logic_item(logic_item)
            check_for_loops(updated_project, logic_item.parse_start().start_action_id)

        if req.dry_run:
            return

        proj.upsert_logic_item(logic_item)

        evt = sevts.p.LogicItemChanged(logic_item)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))

    asyncio.create_task(
        glob.LOCK.write_unlock(
            [item for item in (req.args.start, req.args.end) if item not in to_lock], user_name, True
        )
    )
    return None


async def update_logic_item_cb(req: srpc.p.UpdateLogicItem.Request, ui: WsClient) -> None:
    # TODO lock RPC when used

    scene = glob.LOCK.scene_or_exception()
    proj = glob.LOCK.project_or_exception()

    updated_project = copy.deepcopy(proj)
    updated_logic_item = updated_project.logic_item(req.args.logic_item_id)

    updated_logic_item.start = req.args.start
    updated_logic_item.end = req.args.end
    updated_logic_item.condition = req.args.condition

    check_logic_item(scene, updated_project, updated_logic_item)

    if updated_logic_item.start != updated_logic_item.START:
        check_for_loops(updated_project, updated_logic_item.parse_start().start_action_id)

    if req.dry_run:
        return

    proj.upsert_logic_item(updated_logic_item)

    evt = sevts.p.LogicItemChanged(updated_logic_item)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def remove_logic_item_cb(req: srpc.p.RemoveLogicItem.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    logic_item = proj.logic_item(req.args.logic_item_id)

    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects([logic_item.start, logic_item.end], user_name)
    async with ctx_write_lock(to_lock, user_name):
        # TODO is it necessary to check something here?
        proj.remove_logic_item(req.args.logic_item_id)

        evt = sevts.p.LogicItemChanged(logic_item)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))

    asyncio.create_task(
        glob.LOCK.write_unlock(
            [item for item in (logic_item.start, logic_item.end) if item not in to_lock], user_name, True
        )
    )
    return None


def check_parameter(proj: CachedProject, parameter: common.ProjectParameter) -> None:

    hlp.is_valid_identifier(parameter.name)

    for pparam in proj.parameters:

        if parameter.id == pparam.id:
            continue

        if parameter.name == pparam.name:
            raise Arcor2Exception("Name has to be unique.")

    # TODO check using (constant?) plugin
    from arcor2 import json

    val = json.loads(parameter.value)

    if not isinstance(val, (int, float, str, bool)):
        raise Arcor2Exception("Only basic types are supported so far.")


async def add_project_parameter_cb(req: srpc.p.AddProjectParameter.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()

    async with ctx_write_lock(glob.LOCK.SpecialValues.PROJECT_NAME, glob.USERS.user_name(ui)):

        pparam = common.ProjectParameter(req.args.name, req.args.type, req.args.value)
        check_parameter(proj, pparam)

        if req.dry_run:
            return

        proj.upsert_parameter(pparam)

        evt = sevts.p.ProjectParameterChanged(pparam)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_project_parameter_cb(req: srpc.p.UpdateProjectParameter.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    param = proj.parameter(req.args.id)
    user_name = glob.USERS.user_name(ui)

    await ensure_write_locked(req.args.id, user_name)

    updated_param = copy.deepcopy(param)

    if req.args.name is not None:
        updated_param.name = req.args.name
    if req.args.value is not None:
        updated_param.value = req.args.value

    check_parameter(proj, updated_param)

    if req.dry_run:
        return

    proj.upsert_parameter(updated_param)

    evt = sevts.p.ProjectParameterChanged(updated_param)
    evt.change_type = Event.Type.UPDATE
    asyncio.create_task(notif.broadcast_event(evt))

    asyncio.create_task(glob.LOCK.write_unlock(param.id, user_name, True))


async def remove_project_parameter_cb(req: srpc.p.RemoveProjectParameter.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    pparam = proj.parameter(req.args.id)

    user_name = glob.USERS.user_name(ui)
    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        # check for usage
        for act in proj.actions:
            for param in act.parameters:
                if (
                    param.type == common.ActionParameter.TypeEnum.PROJECT_PARAMETER
                    and param.str_from_value() == pparam.id
                ):
                    raise Arcor2Exception(f"Project parameter used in {act.name} action.")

        if req.dry_run:
            return

        await glob.LOCK.write_unlock(req.args.id, user_name)

        proj.remove_parameter(pparam.id)

        evt = sevts.p.ProjectParameterChanged(pparam)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))


async def delete_project_cb(req: srpc.p.DeleteProject.Request, ui: WsClient) -> None:

    if glob.LOCK.project:
        raise Arcor2Exception("Project has to be closed first.")

    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.id, user_name, auto_unlock=False):
        project = await storage.get_project(req.args.id)
        await glob.LOCK.write_unlock(req.args.id, user_name)
        await storage.delete_project(req.args.id)

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def rename_project_cb(req: srpc.p.RenameProject.Request, ui: WsClient) -> None:

    unique_name(req.args.new_name, (await project_names()))
    user_name = glob.USERS.user_name(ui)

    await ensure_write_locked(req.args.project_id, user_name)

    if req.dry_run:
        return None

    async with managed_project(req.args.project_id) as project:

        project.name = req.args.new_name
        project.update_modified()

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))

    asyncio.create_task(glob.LOCK.write_unlock(req.args.project_id, user_name))
    return None


async def copy_project_cb(req: srpc.p.CopyProject.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.source_id, glob.USERS.user_name(ui)):
        unique_name(req.args.target_name, (await project_names()))

        if req.dry_run:
            return None

        async with managed_project(req.args.source_id, make_copy=True) as project:

            project.name = req.args.target_name

            evt = sevts.p.ProjectChanged(project.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))

        return None


async def update_project_description_cb(req: srpc.p.UpdateProjectDescription.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.project_id, glob.USERS.user_name(ui)):
        async with managed_project(req.args.project_id) as project:

            project.description = req.args.new_description
            project.update_modified()

            evt = sevts.p.ProjectChanged(project.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_project_has_logic_cb(req: srpc.p.UpdateProjectHasLogic.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.project_id, glob.USERS.user_name(ui)):
        async with managed_project(req.args.project_id) as project:

            if project.has_logic and not req.args.new_has_logic:
                project.clear_logic()

                """
                TODO

                if glob.LOCK.project and glob.LOCK.project.id == req.args.project_id:
                ...send remove event for each logic item?
                """

            project.has_logic = req.args.new_has_logic
            project.update_modified()

            evt = sevts.p.ProjectChanged(project.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def rename_action_point_joints_cb(req: srpc.p.RenameActionPointJoints.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    ap, joints = proj.ap_and_joints(req.args.joints_id)

    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, proj.ap_joint_names(ap.id))

    await ensure_write_locked(ap.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    joints.name = req.args.new_name
    proj.update_modified()

    evt = sevts.p.JointsChanged(joints)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None


async def rename_action_point_orientation_cb(req: srpc.p.RenameActionPointOrientation.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    ap, ori = proj.bare_ap_and_orientation(req.args.orientation_id)

    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, proj.ap_orientation_names(ap.id))

    await ensure_write_locked(ori.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    ori.name = req.args.new_name
    proj.update_modified()

    evt = sevts.p.OrientationChanged(ori)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None


async def rename_action_cb(req: srpc.p.RenameAction.Request, ui: WsClient) -> None:

    proj = glob.LOCK.project_or_exception()
    unique_name(req.args.new_name, proj.action_names)
    user_name = glob.USERS.user_name(ui)

    await ensure_write_locked(req.args.action_id, user_name)

    if req.dry_run:
        return None

    act = proj.action(req.args.action_id)
    act.name = req.args.new_name

    proj.update_modified()

    asyncio.create_task(glob.LOCK.write_unlock(req.args.action_id, user_name, True))

    evt = sevts.p.ActionChanged(act)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None



import asyncio
import math
import time
from typing import Awaitable, Callable, Dict, List, Optional, Set

import numpy as np
from arcor2_calibration_data import client as calib_client
from arcor2_calibration_data.client import CalibrateRobotArgs
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import transformations as tr
from arcor2 import ws_server
from arcor2.clients.project_service import URL as ps_url
from arcor2.data import common
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import run_in_executor
from arcor2.object_types.abstract import Camera, MultiArmRobot, Robot
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import robot
from arcor2_arserver.helpers import ctx_read_lock, ctx_write_lock, ensure_write_locked
from arcor2_arserver.scene import (
    ensure_scene_started,
    get_instance,
    get_robot_instance,
    scene_started,
    update_scene_object_pose,
)
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc
from arcor2_arserver_data.events.common import ProcessState
from arcor2_arserver_data.events.robot import HandTeachingMode

RBT_CALIB = "RobotCalibration"

TaskDict = Dict[str, asyncio.Task]

ROBOT_JOINTS_TASKS: TaskDict = {}
EEF_POSE_TASKS: TaskDict = {}

EVENT_PERIOD = 0.1


async def robot_joints_event(robot_inst: Robot) -> None:

    glob.logger.info(f"Sending joints for {robot_inst.name} started.")
    while scene_started() and glob.ROBOT_JOINTS_REGISTERED_UIS[robot_inst.id]:

        start = time.monotonic()

        try:
            evt = sevts.r.RobotJoints(
                sevts.r.RobotJoints.Data(robot_inst.id, (await robot.get_robot_joints(robot_inst, None)))
            )
        except Arcor2Exception as e:
            glob.logger.error(f"Failed to get joints for {robot_inst.name}. {str(e)}")
            break

        evt_json = evt.to_json()
        await asyncio.gather(
            *[ws_server.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_JOINTS_REGISTERED_UIS[robot_inst.id]]
        )

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD - (end - start))

    ROBOT_JOINTS_TASKS.pop(robot_inst.id, None)

    # TODO notify UIs that registration was cancelled
    glob.ROBOT_JOINTS_REGISTERED_UIS.pop(robot_inst.id, None)

    glob.logger.info(f"Sending joints for {robot_inst.name} stopped.")


async def eef_pose(robot_inst: Robot, eef_id: str, arm_id: Optional[str] = None) -> sevts.r.RobotEef.Data.EefPose:

    return sevts.r.RobotEef.Data.EefPose(
        eef_id, (await robot.get_end_effector_pose(robot_inst, eef_id, arm_id)), arm_id
    )


async def robot_eef_pose_event(robot_inst: Robot) -> None:

    eefs: Dict[Optional[str], Set[str]] = {}

    try:
        try:
            for arm_id in await robot.get_arms(robot_inst):
                eefs[arm_id] = await robot.get_end_effectors(robot_inst, arm_id)
        except robot.SingleArmRobotException:
            eefs = {None: await robot.get_end_effectors(robot_inst, None)}
    except Arcor2Exception as e:
        glob.logger.error(f"Failed to start sending poses for {robot_inst.name}. {str(e)}")
    else:

        glob.logger.info(f"Sending poses for {robot_inst.name} started. Arms/EEFs: {eefs}.")

        while scene_started() and glob.ROBOT_EEF_REGISTERED_UIS[robot_inst.id]:

            start = time.monotonic()

            evt = sevts.r.RobotEef(sevts.r.RobotEef.Data(robot_inst.id))

            try:
                evt.data.end_effectors = await asyncio.gather(
                    *[eef_pose(robot_inst, eef_id, arm_id) for arm_id in eefs.keys() for eef_id in eefs[arm_id]]
                )
            except Arcor2Exception as e:
                glob.logger.error(f"Failed to get eef pose for {robot_inst.name}. {str(e)}")
                break

            evt_json = evt.to_json()
            await asyncio.gather(
                *[ws_server.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_EEF_REGISTERED_UIS[robot_inst.id]]
            )

            end = time.monotonic()
            await asyncio.sleep(EVENT_PERIOD - (end - start))

    EEF_POSE_TASKS.pop(robot_inst.id, None)

    # TODO notify UIs that registration was cancelled
    glob.ROBOT_EEF_REGISTERED_UIS.pop(robot_inst.id, None)

    glob.logger.info(f"Sending poses for {robot_inst.name} stopped.")


async def get_robot_meta_cb(req: srpc.r.GetRobotMeta.Request, ui: WsClient) -> srpc.r.GetRobotMeta.Response:

    return srpc.r.GetRobotMeta.Response(
        data=[obj.robot_meta for obj in glob.OBJECT_TYPES.values() if obj.robot_meta is not None]
    )


async def get_robot_joints_cb(req: srpc.r.GetRobotJoints.Request, ui: WsClient) -> srpc.r.GetRobotJoints.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetRobotJoints.Response(
            data=await robot.get_robot_joints(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def get_end_effector_pose_cb(
    req: srpc.r.GetEndEffectorPose.Request, ui: WsClient
) -> srpc.r.GetEndEffectorPose.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetEndEffectorPose.Response(
            data=await robot.get_end_effector_pose(
                get_robot_instance(req.args.robot_id), req.args.end_effector_id, req.args.arm_id
            )
        )


async def get_end_effectors_cb(req: srpc.r.GetEndEffectors.Request, ui: WsClient) -> srpc.r.GetEndEffectors.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetEndEffectors.Response(
            data=await robot.get_end_effectors(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def get_robot_arms_cb(req: srpc.r.GetRobotArms.Request, ui: WsClient) -> srpc.r.GetRobotArms.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetRobotArms.Response(data=await robot.get_arms(get_robot_instance(req.args.robot_id)))


async def get_grippers_cb(req: srpc.r.GetGrippers.Request, ui: WsClient) -> srpc.r.GetGrippers.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetGrippers.Response(
            data=await robot.get_grippers(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def get_suctions_cb(req: srpc.r.GetSuctions.Request, ui: WsClient) -> srpc.r.GetSuctions.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()
        return srpc.r.GetSuctions.Response(
            data=await robot.get_suctions(get_robot_instance(req.args.robot_id), req.args.arm_id)
        )


async def register(
    req: srpc.r.RegisterForRobotEvent.Request,
    robot_inst: Robot,
    ui: WsClient,
    tasks: TaskDict,
    reg_uis: glob.RegisteredUiDict,
    coro: Callable[[Robot], Awaitable[None]],
) -> None:

    if req.args.send:

        reg_uis[req.args.robot_id].add(ui)

        if req.args.robot_id not in tasks:
            # start task
            tasks[req.args.robot_id] = asyncio.create_task(coro(robot_inst))

    else:
        try:
            reg_uis[req.args.robot_id].remove(ui)
        except KeyError as e:
            raise Arcor2Exception("Failed to unregister.") from e

        # cancel task if not needed anymore
        if not reg_uis[req.args.robot_id]:
            task = tasks[req.args.robot_id]

            if not task.cancelled():
                task.cancel()

            del tasks[req.args.robot_id]


async def register_for_robot_event_cb(req: srpc.r.RegisterForRobotEvent.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.robot_id, glob.USERS.user_name(ui)):
        ensure_scene_started()

        # check if robot exists
        robot_inst = get_robot_instance(req.args.robot_id)

        if req.args.what == req.args.RegisterEnum.JOINTS:
            await register(
                req, robot_inst, ui, ROBOT_JOINTS_TASKS, glob.ROBOT_JOINTS_REGISTERED_UIS, robot_joints_event
            )
        elif req.args.what == req.args.RegisterEnum.EEF_POSE:

            if isinstance(robot_inst, MultiArmRobot):
                for arm_id in await robot.get_arms(robot_inst):
                    if await robot.get_end_effectors(robot_inst, arm_id):
                        break
                else:
                    raise Arcor2Exception("Robot does not have any end effector.")

            elif not await robot.get_end_effectors(robot_inst, None):
                raise Arcor2Exception("Robot does not have any end effector.")

            await register(req, robot_inst, ui, EEF_POSE_TASKS, glob.ROBOT_EEF_REGISTERED_UIS, robot_eef_pose_event)
        else:
            raise Arcor2Exception(f"Option '{req.args.what.value}' not implemented.")

        return None


async def check_feature(robot_inst: Robot, feature_name: str) -> None:

    obj_type = glob.OBJECT_TYPES[robot_inst.__class__.__name__]

    if obj_type.robot_meta is None:
        raise Arcor2Exception("Not a robot.")

    try:
        if not getattr(obj_type.robot_meta.features, feature_name):
            raise Arcor2Exception(f"Robot does not support '{feature_name}' feature.")
    except AttributeError:
        raise Arcor2Exception(f"Unknown robot feature: {feature_name}.")


async def move_to_pose_cb(req: srpc.r.MoveToPose.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.robot_id, user_name, auto_unlock=False):
        ensure_scene_started()

        robot_inst = get_robot_instance(req.args.robot_id)
        await robot.check_eef_arm(robot_inst, req.args.arm_id, req.args.end_effector_id)

        await check_feature(robot_inst, Robot.move_to_pose.__name__)
        await robot.check_robot_before_move(robot_inst)

        if (req.args.position is None) != (req.args.orientation is None):

            target_pose = await robot.get_end_effector_pose(robot_inst, req.args.end_effector_id, req.args.arm_id)

            if req.args.position:
                target_pose.position = req.args.position
            elif req.args.orientation:
                target_pose.orientation = req.args.orientation

        elif req.args.position is not None and req.args.orientation is not None:
            target_pose = common.Pose(req.args.position, req.args.orientation)
        else:
            raise Arcor2Exception("Position or orientation should be given.")

        # TODO check if the target pose is reachable (dry_run)
        asyncio.ensure_future(
            robot.move_to_pose(
                robot_inst,
                req.args.end_effector_id,
                req.args.arm_id,
                target_pose,
                req.args.speed,
                req.args.safe,
                user_name,
            )
        )


async def move_to_joints_cb(req: srpc.r.MoveToJoints.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.robot_id, user_name, auto_unlock=False):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)
        await check_feature(robot_inst, Robot.move_to_joints.__name__)
        await robot.check_robot_before_move(robot_inst)

        asyncio.ensure_future(
            robot.move_to_joints(robot_inst, req.args.joints, req.args.speed, req.args.safe, user_name, req.args.arm_id)
        )


async def stop_robot_cb(req: srpc.r.StopRobot.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    # Stop robot cannot use lock, because robot is locked when action is called. Stop will also release lock.
    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)
    await check_feature(robot_inst, Robot.stop.__name__)
    await robot.stop(robot_inst)


async def move_to_action_point_cb(req: srpc.r.MoveToActionPoint.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()
    project = glob.LOCK.project_or_exception()

    async with ctx_write_lock(req.args.robot_id, glob.USERS.user_name(ui)):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)

        await robot.check_robot_before_move(robot_inst)

        if (req.args.orientation_id is None) == (req.args.joints_id is None):
            raise Arcor2Exception("Set orientation or joints. Not both.")

        if req.args.orientation_id:

            await check_feature(robot_inst, Robot.move_to_pose.__name__)

            if req.args.end_effector_id is None:
                raise Arcor2Exception("eef id has to be set.")

            pose = tr.abs_pose_from_ap_orientation(scene, project, req.args.orientation_id)

            # TODO check if the target pose is reachable (dry_run)
            asyncio.ensure_future(
                robot.move_to_ap_orientation(
                    robot_inst,
                    req.args.end_effector_id,
                    req.args.arm_id,
                    pose,
                    req.args.speed,
                    req.args.orientation_id,
                    req.args.safe,
                )
            )

        elif req.args.joints_id:

            await check_feature(robot_inst, Robot.move_to_joints.__name__)

            # TODO add arm_id to ProjectRobotJoints and use it as a parameter for move_to_ap_joints
            joints = project.joints(req.args.joints_id)

            # TODO check if the joints are within limits and reachable (dry_run)
            asyncio.ensure_future(
                robot.move_to_ap_joints(
                    robot_inst, joints.joints, req.args.speed, req.args.joints_id, req.args.safe, req.args.arm_id
                )
            )


async def ik_cb(req: srpc.r.InverseKinematics.Request, ui: WsClient) -> srpc.r.InverseKinematics.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock([req.args.robot_id, req.args.end_effector_id], glob.USERS.user_name(ui)):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)
        await check_feature(robot_inst, Robot.inverse_kinematics.__name__)

        joints = await robot.ik(
            robot_inst,
            req.args.end_effector_id,
            req.args.arm_id,
            req.args.pose,
            req.args.start_joints,
            req.args.avoid_collisions,
        )
        resp = srpc.r.InverseKinematics.Response()
        resp.data = joints
        return resp


async def fk_cb(req: srpc.r.ForwardKinematics.Request, ui: WsClient) -> srpc.r.ForwardKinematics.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock([req.args.robot_id, req.args.end_effector_id], glob.USERS.user_name(ui)):

        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)
        await check_feature(robot_inst, Robot.forward_kinematics.__name__)

        pose = await robot.fk(robot_inst, req.args.end_effector_id, req.args.arm_id, req.args.joints)
        resp = srpc.r.ForwardKinematics.Response()
        resp.data = pose
        return resp


async def calibrate_robot(
    robot_inst: Robot, camera_inst: Camera, move_to_calibration_pose: bool, user_name: str
) -> None:

    try:
        scene = glob.LOCK.scene_or_exception()

        assert camera_inst.color_camera_params

        await notif.broadcast_event(ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Started)))

        try:

            if move_to_calibration_pose:
                await run_in_executor(robot_inst.move_to_calibration_pose)
            robot_joints = await run_in_executor(robot_inst.robot_joints)
            depth_image = await run_in_executor(camera_inst.depth_image, 128)

            args = CalibrateRobotArgs(
                robot_joints,
                robot_inst.pose,
                camera_inst.pose,
                camera_inst.color_camera_params,
                f"{ps_url}/models/{robot_inst.urdf_package_name}/mesh/file",
            )

            new_pose = await run_in_executor(calib_client.calibrate_robot, args, depth_image)

        except Arcor2Exception as e:
            await notif.broadcast_event(
                ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Failed, str(e)))
            )
            glob.logger.exception("Failed to calibrate the robot.")
            return

        await update_scene_object_pose(scene, scene.object(robot_inst.id), new_pose, robot_inst)
        await notif.broadcast_event(ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Finished)))
    finally:
        await glob.LOCK.write_unlock(camera_inst.id, user_name)


async def calibrate_robot_cb(req: srpc.r.CalibrateRobot.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()
    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.camera_id, user_name, auto_unlock=False):
        ensure_scene_started()
        robot_inst = get_robot_instance(req.args.robot_id)

        if not robot_inst.urdf_package_name:
            raise Arcor2Exception("Robot with model required!")

        if req.args.camera_id:
            camera_inst = get_instance(req.args.camera_id, Camera)
        else:
            for obj in glob.SCENE_OBJECT_INSTANCES.values():
                if isinstance(obj, Camera):
                    camera_inst = obj
                    break
            else:
                raise Arcor2Exception("No camera found.")

        if camera_inst.color_camera_params is None:
            raise Arcor2Exception("Calibrated camera required!")

        await ensure_write_locked(req.args.robot_id, user_name)

        asyncio.ensure_future(calibrate_robot(robot_inst, camera_inst, req.args.move_to_calibration_pose, user_name))

        return None


async def hand_teaching_mode_cb(req: srpc.r.HandTeachingMode.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)

    # in this case, method name does not correspond to feature name
    await check_feature(robot_inst, "hand_teaching")

    hand_teaching_mode = await run_in_executor(robot_inst.get_hand_teaching_mode, req.args.arm_id)

    if req.args.enable == hand_teaching_mode:
        raise Arcor2Exception("That's the current state.")

    await ensure_write_locked(req.args.robot_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    await run_in_executor(robot_inst.set_hand_teaching_mode, req.args.enable, req.args.arm_id)
    evt = HandTeachingMode(HandTeachingMode.Data(req.args.robot_id, req.args.enable, req.args.arm_id))
    asyncio.ensure_future(notif.broadcast_event(evt))


async def step_robot_eef_cb(req: srpc.r.StepRobotEef.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()

    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)

    await check_feature(robot_inst, Robot.move_to_pose.__name__)
    await robot.check_robot_before_move(robot_inst)

    tp = await robot.get_end_effector_pose(robot_inst, req.args.end_effector_id, req.args.arm_id)

    if req.args.mode == req.args.mode.ROBOT:
        tp = tr.make_pose_rel(robot_inst.pose, tp)
    elif req.args.mode == req.args.mode.RELATIVE:
        assert req.args.pose
        tp = tr.make_pose_rel(req.args.pose, tp)
    elif req.args.mode == req.args.mode.USER:
        assert req.args.pose
        raise Arcor2Exception("Not supported yet.")

    if req.args.what == req.args.what.POSITION:
        if req.args.axis == req.args.axis.X:
            tp.position.x += req.args.step
        elif req.args.axis == req.args.axis.Y:
            tp.position.y += req.args.step
        elif req.args.axis == req.args.axis.Z:
            tp.position.z += req.args.step
    elif req.args.what == req.args.what.ORIENTATION:
        if req.args.axis == req.args.axis.X:
            tp.orientation *= common.Orientation.from_rotation_vector(x=req.args.step)
        elif req.args.axis == req.args.axis.Y:
            tp.orientation *= common.Orientation.from_rotation_vector(y=req.args.step)
        elif req.args.axis == req.args.axis.Z:
            tp.orientation *= common.Orientation.from_rotation_vector(z=req.args.step)

    if req.args.mode == req.args.mode.ROBOT:
        tp = tr.make_pose_abs(robot_inst.pose, tp)
    elif req.args.mode == req.args.mode.RELATIVE:
        assert req.args.pose
        tp = tr.make_pose_abs(req.args.pose, tp)

    await robot.check_reachability(scene, robot_inst, req.args.end_effector_id, req.args.arm_id, tp, req.args.safe)

    await ensure_write_locked(req.args.robot_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    asyncio.ensure_future(
        robot.move_to_pose(robot_inst, req.args.end_effector_id, req.args.arm_id, tp, req.args.speed, req.args.safe)
    )


async def set_eef_perpendicular_to_world_cb(req: srpc.r.SetEefPerpendicularToWorld.Request, ui: WsClient) -> None:

    glob.LOCK.scene_or_exception()

    ensure_scene_started()
    robot_inst = get_robot_instance(req.args.robot_id)

    await check_feature(robot_inst, Robot.move_to_pose.__name__)
    await check_feature(robot_inst, Robot.inverse_kinematics.__name__)
    await robot.check_robot_before_move(robot_inst)

    await ensure_write_locked(req.args.robot_id, glob.USERS.user_name(ui))

    if req.dry_run:  # attempt to find suitable joints can take some time so it is not done for dry_run
        return

    tp, current_joints = await robot.get_pose_and_joints(robot_inst, req.args.end_effector_id, req.args.arm_id)

    target_joints: Optional[List[common.Joint]] = None
    target_joints_diff: float = 0.0

    # select best (closest joint configuration) reachable pose
    tasks = [
        robot.ik(robot_inst, req.args.end_effector_id, req.args.arm_id, pose, current_joints, req.args.safe)
        for pose in [
            common.Pose(
                tp.position,
                common.Orientation.from_rotation_vector(y=math.pi) * common.Orientation.from_rotation_vector(z=z_rot),
            )
            for z_rot in np.linspace(-math.pi, math.pi, 360)
        ]
    ]

    for res in await asyncio.gather(*tasks, return_exceptions=True):

        if not isinstance(res, list):
            continue

        if not target_joints:
            target_joints = res
            for f, b in zip(current_joints, target_joints):
                assert f.name == b.name
                target_joints_diff += (f.value - b.value) ** 2
        else:
            diff = 0.0
            for f, b in zip(current_joints, res):
                assert f.name == b.name
                diff += (f.value - b.value) ** 2

            if diff < target_joints_diff:
                target_joints = res
                target_joints_diff = diff

    if not target_joints:
        raise Arcor2Exception("Could not find reachable pose.")

    asyncio.ensure_future(
        robot.move_to_joints(robot_inst, target_joints, req.args.speed, req.args.safe, req.args.arm_id)
    )



import asyncio
import inspect
from typing import Any, List, Optional, Set, Tuple, Type

import arcor2.helpers as hlp
from arcor2.cached import CachedScene
from arcor2.data import common
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import MultiArmRobot, Robot
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import objects_actions as osa
from arcor2_arserver.clients import project_service as ps
from arcor2_arserver.object_types.data import ObjectTypeData
from arcor2_arserver.object_types.source import function_implemented
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data.robot import RobotMeta


class RobotPoseException(Arcor2Exception):
    pass


class SingleArmRobotException(Arcor2Exception):
    pass


def prepare_args(robot_inst: Robot, args: List[Any], arm_id: Optional[str]) -> List[Any]:

    if isinstance(robot_inst, MultiArmRobot):
        args.append(arm_id)
    elif arm_id:
        raise SingleArmRobotException("Single arm robot.")

    return args


async def get_arms(robot_inst: Robot) -> Set[str]:
    """
    :param robot_inst:
    :return: IDs of existing arms.
    """

    if not isinstance(robot_inst, MultiArmRobot):
        raise SingleArmRobotException(f"{robot_inst.__class__.__name__} has only one arm.")

    return await hlp.run_in_executor(robot_inst.get_arm_ids)


async def get_end_effectors(robot_inst: Robot, arm_id: Optional[str]) -> Set[str]:
    """
    :param robot_inst:
    :return: IDs of existing end effectors.
    """

    return await hlp.run_in_executor(robot_inst.get_end_effectors_ids, *prepare_args(robot_inst, [], arm_id))


async def get_grippers(robot_inst: Robot, arm_id: Optional[str]) -> Set[str]:
    """
    :param robot_inst:
    :return: IDs of existing grippers.
    """

    return await hlp.run_in_executor(robot_inst.grippers, *prepare_args(robot_inst, [], arm_id))


async def get_suctions(robot_inst: Robot, arm_id: Optional[str]) -> Set[str]:
    """
    :param robot_inst:
    :return: IDs of existing suctions.
    """

    return await hlp.run_in_executor(robot_inst.suctions, *prepare_args(robot_inst, [], arm_id))


async def get_pose_and_joints(
    robot_inst: Robot, end_effector: str, arm_id: Optional[str]
) -> Tuple[common.Pose, List[common.Joint]]:

    return await asyncio.gather(
        hlp.run_in_executor(robot_inst.get_end_effector_pose, *prepare_args(robot_inst, [end_effector], arm_id)),
        hlp.run_in_executor(robot_inst.robot_joints, *prepare_args(robot_inst, [], arm_id)),
    )


async def get_end_effector_pose(robot_inst: Robot, end_effector: str, arm_id: Optional[str]) -> common.Pose:
    """
    :param robot_inst:
    :param end_effector:
    :return: Global pose
    """

    return await hlp.run_in_executor(
        robot_inst.get_end_effector_pose, *prepare_args(robot_inst, [end_effector], arm_id)
    )


async def get_robot_joints(robot_inst: Robot, arm_id: Optional[str]) -> List[common.Joint]:
    """
    :param robot_inst:
    :return: List of joints
    """

    return await hlp.run_in_executor(robot_inst.robot_joints, *prepare_args(robot_inst, [], arm_id))


def _feature(type_def: Type[Robot], method_name: str, base_class: Type[Robot]) -> bool:

    assert glob.OBJECT_TYPES

    method = getattr(type_def, method_name)
    where_it_is_defined = glob.OBJECT_TYPES.get(method.__qualname__.split(".")[0], None)

    if where_it_is_defined is None:
        raise Arcor2Exception(f"Can't get origin for {type_def.__name__}/{method_name}.")

    if where_it_is_defined.type_def is None or where_it_is_defined.ast is None:
        raise Arcor2Exception(
            f"Origin {where_it_is_defined.meta.type} for {type_def.__name__}/{method_name} is disabled."
        )

    glob.logger.debug(
        f"Processing {type_def.__name__}/{method_name} "
        f"(defined in {where_it_is_defined.type_def.__name__}), with base class {base_class.__name__}."
    )

    if where_it_is_defined.type_def is base_class or where_it_is_defined.meta.disabled:
        # all of the "feature" methods are abstract in the base class and have to be implemented for the concrete robot
        return False

    if not function_implemented(where_it_is_defined.ast, method_name):
        glob.logger.debug(f"{type_def.__name__}/{method_name} not implemented.")
        return False

    sign = inspect.signature(getattr(where_it_is_defined.type_def, method_name))
    if not (res := inspect.signature(getattr(base_class, method_name)) == sign):
        glob.logger.debug(f"{type_def.__name__}/{method_name} has invalid signature.")
    return res


async def get_robot_meta(obj_type: ObjectTypeData) -> None:

    if obj_type.meta.disabled:
        raise Arcor2Exception("Disabled object type.")

    assert obj_type.type_def is not None

    if not issubclass(obj_type.type_def, Robot):
        raise Arcor2Exception("Not a robot.")

    obj_type.robot_meta = RobotMeta(
        obj_type.meta.type, obj_type.type_def.robot_type, issubclass(obj_type.type_def, MultiArmRobot)
    )

    # TODO fix mypy issue 'Can only assign concrete classes to a variable of type "Type[Robot]"'
    base_class: Type[Robot] = MultiArmRobot if obj_type.robot_meta.multi_arm else Robot  # type: ignore

    # TODO automate this somehow
    obj_type.robot_meta.features.move_to_pose = _feature(obj_type.type_def, Robot.move_to_pose.__name__, base_class)
    obj_type.robot_meta.features.move_to_joints = _feature(obj_type.type_def, Robot.move_to_joints.__name__, base_class)
    obj_type.robot_meta.features.stop = _feature(obj_type.type_def, Robot.stop.__name__, base_class)
    obj_type.robot_meta.features.inverse_kinematics = _feature(
        obj_type.type_def, Robot.inverse_kinematics.__name__, base_class
    )
    obj_type.robot_meta.features.forward_kinematics = _feature(
        obj_type.type_def, Robot.forward_kinematics.__name__, base_class
    )
    obj_type.robot_meta.features.hand_teaching = _feature(
        obj_type.type_def, Robot.set_hand_teaching_mode.__name__, base_class
    )

    if urdf_name := obj_type.type_def.urdf_package_name:

        if urdf_name not in await ps.files_ids():
            glob.logger.error(f"URDF package {urdf_name} for {obj_type.meta.type} does not exist.")
        else:
            obj_type.robot_meta.urdf_package_filename = urdf_name
            # TODO check if URDF is valid?

    glob.logger.debug(obj_type.robot_meta)


async def stop(robot_inst: Robot) -> None:

    if not robot_inst.move_in_progress:
        raise Arcor2Exception("Robot is not moving.")

    await hlp.run_in_executor(robot_inst.stop)


async def ik(
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: Optional[str],
    pose: common.Pose,
    start_joints: Optional[List[common.Joint]] = None,
    avoid_collisions: bool = True,
) -> List[common.Joint]:

    return await hlp.run_in_executor(
        robot_inst.inverse_kinematics,
        *prepare_args(robot_inst, [end_effector_id, pose, start_joints, avoid_collisions], arm_id),
    )


async def fk(robot_inst: Robot, end_effector_id: str, arm_id: Optional[str], joints: List[common.Joint]) -> common.Pose:

    return await hlp.run_in_executor(
        robot_inst.forward_kinematics, *prepare_args(robot_inst, [end_effector_id, joints], arm_id)
    )


async def check_reachability(
    scene: CachedScene,
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: Optional[str],
    pose: common.Pose,
    safe: bool = True,
) -> None:

    otd = osa.get_obj_type_data(scene, robot_inst.id)
    if otd.robot_meta and otd.robot_meta.features.inverse_kinematics:
        try:
            await ik(robot_inst, end_effector_id, arm_id, pose, avoid_collisions=safe)
        except Arcor2Exception as e:
            raise Arcor2Exception("Unreachable pose.") from e


async def check_robot_before_move(robot_inst: Robot) -> None:

    robot_inst.check_if_ready_to_move()

    if glob.RUNNING_ACTION:

        assert glob.LOCK.project
        action = glob.LOCK.project.action(glob.RUNNING_ACTION)
        obj_id_str, _ = action.parse_type()

        if robot_inst.id == obj_id_str:
            raise Arcor2Exception("Robot is executing action.")


async def _move_to_pose(
    robot_inst: Robot, end_effector_id: str, arm_id: Optional[str], pose: common.Pose, speed: float, safe: bool
) -> None:
    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    try:
        await hlp.run_in_executor(
            robot_inst.move_to_pose, *prepare_args(robot_inst, [end_effector_id, pose, speed, safe], arm_id)
        )
    except Arcor2Exception as e:
        glob.logger.error(f"Robot movement failed with: {str(e)}")
        raise


async def move_to_pose(
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: Optional[str],
    pose: common.Pose,
    speed: float,
    safe: bool,
    lock_owner: Optional[str] = None,
) -> None:

    try:
        Data = sevts.r.RobotMoveToPose.Data

        await notif.broadcast_event(
            sevts.r.RobotMoveToPose(
                Data(Data.MoveEventType.START, robot_inst.id, end_effector_id, pose, safe, arm_id=arm_id)
            )
        )

        try:
            await _move_to_pose(robot_inst, end_effector_id, arm_id, pose, speed, safe)

        except Arcor2Exception as e:
            await notif.broadcast_event(
                sevts.r.RobotMoveToPose(
                    Data(Data.MoveEventType.FAILED, robot_inst.id, end_effector_id, pose, safe, str(e), arm_id)
                )
            )
            return

        await notif.broadcast_event(
            sevts.r.RobotMoveToPose(
                Data(Data.MoveEventType.END, robot_inst.id, end_effector_id, pose, safe, arm_id=arm_id)
            )
        )
    finally:
        if lock_owner:
            await glob.LOCK.write_unlock(robot_inst.id, lock_owner)


async def move_to_ap_orientation(
    robot_inst: Robot,
    end_effector_id: str,
    arm_id: Optional[str],
    pose: common.Pose,
    speed: float,
    orientation_id: str,
    safe: bool,
) -> None:

    Data = sevts.r.RobotMoveToActionPointOrientation.Data

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointOrientation(
            Data(Data.MoveEventType.START, robot_inst.id, end_effector_id, orientation_id, safe, arm_id=arm_id)
        )
    )

    try:
        await _move_to_pose(robot_inst, end_effector_id, arm_id, pose, speed, safe)

    except Arcor2Exception as e:
        await notif.broadcast_event(
            sevts.r.RobotMoveToActionPointOrientation(
                Data(Data.MoveEventType.FAILED, robot_inst.id, end_effector_id, orientation_id, safe, str(e), arm_id)
            )
        )
        return

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointOrientation(
            Data(Data.MoveEventType.END, robot_inst.id, end_effector_id, orientation_id, safe, arm_id=arm_id)
        )
    )


async def _move_to_joints(
    robot_inst: Robot, joints: List[common.Joint], speed: float, safe: bool, arm_id: Optional[str]
) -> None:

    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    try:
        await hlp.run_in_executor(robot_inst.move_to_joints, *prepare_args(robot_inst, [joints, speed, safe], arm_id))
    except Arcor2Exception as e:
        glob.logger.error(f"Robot movement failed with: {str(e)}")
        raise


async def move_to_joints(
    robot_inst: Robot,
    joints: List[common.Joint],
    speed: float,
    safe: bool,
    arm_id: Optional[str],
    lock_owner: Optional[str] = None,
) -> None:

    try:
        Data = sevts.r.RobotMoveToJoints.Data

        await notif.broadcast_event(
            sevts.r.RobotMoveToJoints(Data(Data.MoveEventType.START, robot_inst.id, joints, safe, arm_id=arm_id))
        )

        try:

            await _move_to_joints(robot_inst, joints, speed, safe, arm_id)

        except Arcor2Exception as e:

            await notif.broadcast_event(
                sevts.r.RobotMoveToJoints(
                    Data(Data.MoveEventType.FAILED, robot_inst.id, joints, safe, str(e), arm_id=arm_id)
                )
            )

            return

        await notif.broadcast_event(
            sevts.r.RobotMoveToJoints(Data(Data.MoveEventType.END, robot_inst.id, joints, safe, arm_id=arm_id))
        )
    finally:
        if lock_owner:
            await glob.LOCK.write_unlock(robot_inst.id, lock_owner)


async def move_to_ap_joints(
    robot_inst: Robot,
    joints: List[common.Joint],
    speed: float,
    joints_id: str,
    safe: bool,
    arm_id: Optional[str],
) -> None:

    Data = sevts.r.RobotMoveToActionPointJoints.Data

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointJoints(
            Data(Data.MoveEventType.START, robot_inst.id, joints_id, safe, arm_id=arm_id)
        )
    )

    try:

        await _move_to_joints(robot_inst, joints, speed, safe, arm_id)

    except Arcor2Exception as e:

        await notif.broadcast_event(
            sevts.r.RobotMoveToActionPointJoints(
                Data(Data.MoveEventType.FAILED, robot_inst.id, joints_id, safe, str(e), arm_id)
            )
        )

        return

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointJoints(
            Data(Data.MoveEventType.END, robot_inst.id, joints_id, safe, arm_id=arm_id)
        )
    )


async def check_eef_arm(robot_inst: Robot, arm_id: Optional[str], eef_id: Optional[str] = None) -> None:

    if isinstance(robot_inst, MultiArmRobot):
        if not arm_id:
            raise Arcor2Exception("Arm has to be specified.")

        if eef_id is not None and eef_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids, arm_id):
            raise Arcor2Exception("Unknown end effector.")

        return

    if arm_id:
        raise Arcor2Exception("Arm should not be specified.")

    if eef_id is not None and eef_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids):
        raise Arcor2Exception("Unknown end effector.")



import asyncio
import copy
import functools
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Optional

import quaternion
from arcor2_calibration_data import client as calibration
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2.cached import UpdateableCachedScene
from arcor2.data import common, object_type
from arcor2.data.events import Event, PackageState
from arcor2.exceptions import Arcor2Exception
from arcor2.image import image_from_str
from arcor2.object_types.abstract import Generic
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver.clients import project_service as storage
from arcor2_arserver.helpers import (
    ctx_read_lock,
    ctx_write_lock,
    ensure_write_locked,
    get_unlocked_objects,
    unique_name,
)
from arcor2_arserver.lock.exceptions import LockingException
from arcor2_arserver.objects_actions import get_object_types
from arcor2_arserver.project import (
    associated_projects,
    invalidate_joints_using_object_as_parent,
    projects_using_object,
    remove_object_references_from_projects,
)
from arcor2_arserver.robot import check_eef_arm, get_end_effector_pose
from arcor2_arserver.scene import (
    add_object_to_scene,
    can_modify_scene,
    check_object_parameters,
    ensure_scene_started,
    get_instance,
    get_robot_instance,
    get_scene_state,
    notify_scene_closed,
    notify_scene_opened,
    open_scene,
    scene_names,
    scene_started,
    scenes,
    start_scene,
    stop_scene,
    update_scene_object_pose,
)
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc


@asynccontextmanager
async def managed_scene(scene_id: str, make_copy: bool = False) -> AsyncGenerator[UpdateableCachedScene, None]:

    save_back = False

    if glob.LOCK.scene and glob.LOCK.scene.id == scene_id:
        if make_copy:
            scene = copy.deepcopy(glob.LOCK.scene)
            save_back = True
        else:
            scene = glob.LOCK.scene
    else:
        save_back = True
        scene = UpdateableCachedScene(await storage.get_scene(scene_id))

    if make_copy:
        scene.id = common.Scene.uid()

    try:
        yield scene
    finally:
        if save_back:
            asyncio.ensure_future(storage.update_scene(scene))


async def new_scene_cb(req: srpc.s.NewScene.Request, ui: WsClient) -> None:
    """Creates and opens a new scene on the server. Fails if any scene is open
    or if scene id/name already exists.

    :param req:
    :return:
    """

    async with glob.LOCK.get_lock(dry_run=req.dry_run):
        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't create scene while package runs.")

        if glob.LOCK.scene:
            raise Arcor2Exception("Scene has to be closed first.")

        for scene_id in await storage.get_scenes():
            if req.args.name == scene_id.name:
                raise Arcor2Exception("Name already used.")

        if req.dry_run:
            return None

        # TODO get_object_types - not ideal, may take quite long time
        #  ...maybe creating a scene should be done similarly to starting one (process with events about its state)?
        await get_object_types()

        glob.LOCK.scene = UpdateableCachedScene(common.Scene(req.args.name, description=req.args.description))
        asyncio.ensure_future(notify_scene_opened(sevts.s.OpenScene(sevts.s.OpenScene.Data(glob.LOCK.scene.scene))))

        return None


async def close_scene_cb(req: srpc.s.CloseScene.Request, ui: WsClient) -> None:
    """Closes scene on the server.

    :param req:
    :return:
    """

    async with ctx_write_lock(glob.LOCK.SpecialValues.SCENE_NAME, glob.USERS.user_name(ui), dry_run=req.dry_run):

        scene = glob.LOCK.scene_or_exception()

        if glob.LOCK.project:
            raise Arcor2Exception("Project has to be closed first.")

        if not req.args.force and scene.has_changes:
            raise Arcor2Exception("Scene has unsaved changes.")

        can_modify_scene()  # can't close scene while started

        if await glob.LOCK.get_locked_roots_count() > 1:
            raise LockingException(glob.LOCK.ErrMessages.SOMETHING_LOCKED.value)

        if req.dry_run:
            return None

        scene_id = scene.id
        glob.LOCK.scene = None
        glob.OBJECTS_WITH_UPDATED_POSE.clear()
        asyncio.ensure_future(notify_scene_closed(scene_id))


async def save_scene_cb(req: srpc.s.SaveScene.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock(req.dry_run):

        scene = glob.LOCK.scene_or_exception()

        if glob.LOCK.project:
            raise Arcor2Exception("Project has to be closed first.")

        if req.dry_run:
            return None

        scene.modified = await storage.update_scene(scene)
        asyncio.ensure_future(notif.broadcast_event(sevts.s.SceneSaved()))
        for obj_id in glob.OBJECTS_WITH_UPDATED_POSE:
            asyncio.ensure_future(invalidate_joints_using_object_as_parent(scene.object(obj_id)))
        glob.OBJECTS_WITH_UPDATED_POSE.clear()
        return None


async def open_scene_cb(req: srpc.s.OpenScene.Request, ui: WsClient) -> None:

    async with glob.LOCK.get_lock():
        if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
            raise Arcor2Exception("Can't open scene while package runs.")

        if glob.LOCK.scene:
            raise Arcor2Exception("Scene already opened.")

        await open_scene(req.args.id)

        assert glob.LOCK.scene
        assert not glob.LOCK.scene.has_changes
        asyncio.ensure_future(
            notify_scene_opened(sevts.s.OpenScene(data=sevts.s.OpenScene.Data(glob.LOCK.scene.scene)))
        )
        return None


async def list_scenes_cb(req: srpc.s.ListScenes.Request, ui: WsClient) -> srpc.s.ListScenes.Response:

    resp = srpc.s.ListScenes.Response()
    resp.data = []

    async for scene in scenes():
        assert scene.created
        assert scene.modified
        resp.data.append(resp.Data(scene.id, scene.name, scene.created, scene.modified, scene.description))

    return resp


async def add_object_to_scene_cb(req: srpc.s.AddObjectToScene.Request, ui: WsClient) -> None:

    async with ctx_write_lock(glob.LOCK.SpecialValues.SCENE_NAME, glob.USERS.user_name(ui)):

        scene = glob.LOCK.scene_or_exception()

        if glob.LOCK.project:
            raise Arcor2Exception("Project has to be closed first.")

        can_modify_scene()

        obj = common.SceneObject(req.args.name, req.args.type, req.args.pose, req.args.parameters)

        await add_object_to_scene(scene, obj, dry_run=req.dry_run)

        if req.dry_run:
            return None

        scene.update_modified()

        evt = sevts.s.SceneObjectChanged(obj)
        evt.change_type = Event.Type.ADD
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def update_object_parameters_cb(req: srpc.s.UpdateObjectParameters.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)

    can_modify_scene()

    obj = scene.object(req.args.id)

    if obj.type not in glob.OBJECT_TYPES:
        raise Arcor2Exception("Unknown object type.")

    obj_type = glob.OBJECT_TYPES[obj.type]

    check_object_parameters(obj_type, req.args.parameters)

    await ensure_write_locked(req.args.id, glob.USERS.user_name(ui))

    if req.dry_run:
        return None

    obj.parameters = req.args.parameters
    scene.update_modified()

    evt = sevts.s.SceneObjectChanged(obj)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def scene_object_usage_request_cb(
    req: srpc.s.SceneObjectUsage.Request, ui: WsClient
) -> srpc.s.SceneObjectUsage.Response:
    """Works for both services and objects.

    :param req:
    :return:
    """

    scene = glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.id, glob.USERS.user_name(ui)):
        if not (any(obj.id == req.args.id for obj in scene.objects)):
            raise Arcor2Exception("Unknown ID.")

        resp = srpc.s.SceneObjectUsage.Response()
        resp.data = set()

        async for project in projects_using_object(scene.id, req.args.id):
            resp.data.add(project.id)

        return resp


# TODO move to objects
async def action_param_values_cb(
    req: srpc.o.ActionParamValues.Request, ui: WsClient
) -> srpc.o.ActionParamValues.Response:

    glob.LOCK.scene_or_exception()

    async with ctx_read_lock(req.args.id, glob.USERS.user_name(ui)):
        ensure_scene_started()

        inst = get_instance(req.args.id, Generic)

        parent_params = {}

        for pp in req.args.parent_params:
            parent_params[pp.id] = pp.value

        try:
            method_name, required_parent_params = inst.DYNAMIC_PARAMS[req.args.param_id]
        except KeyError:
            raise Arcor2Exception("Unknown parameter or values not constrained.")

        if parent_params.keys() != required_parent_params:
            raise Arcor2Exception("Not all required parent params were given.")

        # TODO validate method parameters vs parent_params (check types)?

        resp = srpc.o.ActionParamValues.Response()

        try:
            method = getattr(inst, method_name)
        except AttributeError:
            glob.logger.error(
                f"Unable to get values for parameter {req.args.param_id}, "
                f"object/service {inst.id} has no method named {method_name}."
            )
            raise Arcor2Exception("System error.")

        # TODO update hlp.run_in_executor to support kwargs
        resp.data = await asyncio.get_event_loop().run_in_executor(None, functools.partial(method, **parent_params))
        return resp


async def remove_from_scene_cb(req: srpc.s.RemoveFromScene.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)
    user_name = glob.USERS.user_name(ui)

    to_lock = await get_unlocked_objects(req.args.id, user_name)
    async with ctx_write_lock(to_lock, user_name, auto_unlock=req.dry_run):

        can_modify_scene()

        if not req.args.force and {proj.name async for proj in projects_using_object(scene.id, req.args.id)}:
            raise Arcor2Exception("Can't remove object that is used in project(s).")

        if req.dry_run:
            return None

        if req.args.id not in scene.object_ids:
            raise Arcor2Exception("Unknown id.")

        await glob.LOCK.write_unlock(req.args.id, user_name)

        obj = scene.object(req.args.id)
        scene.delete_object(req.args.id)

        if req.args.id in glob.OBJECTS_WITH_UPDATED_POSE:
            glob.OBJECTS_WITH_UPDATED_POSE.remove(req.args.id)

        evt = sevts.s.SceneObjectChanged(obj)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))

        # TODO this should be done after scene is saved
        asyncio.ensure_future(remove_object_references_from_projects(req.args.id))
        return None


async def update_object_pose_using_robot_cb(req: srpc.o.UpdateObjectPoseUsingRobot.Request, ui: WsClient) -> None:
    """Updates object's pose using a pose of the robot's end effector.

    :param req:
    :return:
    """

    if req.args.id == req.args.robot.robot_id:
        raise Arcor2Exception("Robot cannot update its own pose.")

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)
    user_name = glob.USERS.user_name(ui)

    to_lock = await get_unlocked_objects([obj for obj in (req.args.robot.robot_id, req.args.id)], user_name)
    async with ctx_write_lock(to_lock, user_name):
        ensure_scene_started()

        robot_inst = get_robot_instance(req.args.robot.robot_id)
        await check_eef_arm(robot_inst, req.args.robot.arm_id, req.args.robot.end_effector)

        scene_object = scene.object(req.args.id)

        obj_type = glob.OBJECT_TYPES[scene_object.type]

        if not obj_type.meta.has_pose:
            raise Arcor2Exception("Object without pose.")

        object_model = obj_type.meta.object_model

        if object_model:
            collision_model = object_model.model()
            if isinstance(collision_model, object_type.Mesh) and req.args.pivot != req.args.PivotEnum.MIDDLE:
                raise Arcor2Exception("Only middle pivot point is supported for objects with mesh collision model.")
        elif req.args.pivot != req.args.PivotEnum.MIDDLE:
            raise Arcor2Exception("Only middle pivot point is supported for objects without collision model.")

        new_pose = await get_end_effector_pose(robot_inst, req.args.robot.end_effector, req.args.robot.arm_id)

        position_delta = common.Position()

        if object_model:
            collision_model = object_model.model()
            if isinstance(collision_model, object_type.Box):
                if req.args.pivot == req.args.PivotEnum.TOP:
                    position_delta.z -= collision_model.size_z / 2
                elif req.args.pivot == req.args.PivotEnum.BOTTOM:
                    position_delta.z += collision_model.size_z / 2
            elif isinstance(collision_model, object_type.Cylinder):
                if req.args.pivot == req.args.PivotEnum.TOP:
                    position_delta.z -= collision_model.height / 2
                elif req.args.pivot == req.args.PivotEnum.BOTTOM:
                    position_delta.z += collision_model.height / 2
            elif isinstance(collision_model, object_type.Sphere):
                if req.args.pivot == req.args.PivotEnum.TOP:
                    position_delta.z -= collision_model.radius / 2
                elif req.args.pivot == req.args.PivotEnum.BOTTOM:
                    position_delta.z += collision_model.radius / 2

        position_delta = position_delta.rotated(new_pose.orientation)

        assert scene_object.pose

        scene_object.pose.position = new_pose.position - position_delta

        scene_object.pose.orientation.set_from_quaternion(
            new_pose.orientation.as_quaternion() * quaternion.quaternion(0, 1, 0, 0)
        )

        asyncio.ensure_future(update_scene_object_pose(scene, scene_object))
        return None


async def update_object_pose_cb(req: srpc.s.UpdateObjectPose.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)

    if scene_started():
        try:
            get_robot_instance(req.args.object_id)
        except Arcor2Exception:
            pass  # ok, it should not be a robot
        else:
            raise Arcor2Exception("Robot's pose can be only updated offline.")

    obj = scene.object(req.args.object_id)

    if not obj.pose:
        raise Arcor2Exception("Object without pose.")

    await ensure_write_locked(req.args.object_id, glob.USERS.user_name(ui))

    if req.dry_run:
        return

    asyncio.ensure_future(update_scene_object_pose(scene, obj, req.args.pose))
    return None


async def rename_object_cb(req: srpc.s.RenameObject.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception(ensure_project_closed=True)
    target_obj = scene.object(req.args.id)

    if target_obj.name == req.args.new_name:
        raise Arcor2Exception("Name unchanged")

    for obj_name in scene.object_names():
        if obj_name == req.args.new_name:
            raise Arcor2Exception("Object name already exists.")

    hlp.is_valid_identifier(req.args.new_name)

    user_name = glob.USERS.user_name(ui)
    await ensure_write_locked(req.args.id, user_name)

    if req.dry_run:
        return None

    target_obj.name = req.args.new_name

    scene.update_modified()

    evt = sevts.s.SceneObjectChanged(target_obj)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))

    asyncio.create_task(glob.LOCK.write_unlock(req.args.id, user_name, True))
    return None


async def rename_scene_cb(req: srpc.s.RenameScene.Request, ui: WsClient) -> None:

    unique_name(req.args.new_name, (await scene_names()))

    user_name = glob.USERS.user_name(ui)
    await ensure_write_locked(req.args.id, user_name)

    if req.dry_run:
        return None

    async with managed_scene(req.args.id) as scene:
        scene.name = req.args.new_name

        evt = sevts.s.SceneChanged(scene.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))

    asyncio.create_task(glob.LOCK.write_unlock(req.args.id, user_name, True))
    return None


async def delete_scene_cb(req: srpc.s.DeleteScene.Request, ui: WsClient) -> Optional[srpc.s.DeleteScene.Response]:

    if glob.LOCK.scene:
        raise Arcor2Exception("Scene has to be closed first.")

    user_name = glob.USERS.user_name(ui)

    async with ctx_write_lock(req.args.id, user_name, auto_unlock=req.dry_run):

        if assoc_projects := await associated_projects(req.args.id):
            resp = srpc.s.DeleteScene.Response(result=False)
            resp.messages = ["Scene has associated projects."]
            resp.data = assoc_projects
            asyncio.create_task(glob.LOCK.write_unlock(req.args.id, user_name))
            return resp

        if req.dry_run:
            return None

        scene = UpdateableCachedScene(await storage.get_scene(req.args.id))
        asyncio.create_task(glob.LOCK.write_unlock(req.args.id, user_name))
        await storage.delete_scene(req.args.id)
        evt = sevts.s.SceneChanged(scene.bare)
        evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def projects_with_scene_cb(
    req: srpc.s.ProjectsWithScene.Request, ui: WsClient
) -> srpc.s.ProjectsWithScene.Response:

    async with ctx_read_lock(req.args.id, glob.USERS.user_name(ui)):
        resp = srpc.s.ProjectsWithScene.Response()
        resp.data = await associated_projects(req.args.id)
        return resp


async def update_scene_description_cb(req: srpc.s.UpdateSceneDescription.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.scene_id, glob.USERS.user_name(ui)):
        async with managed_scene(req.args.scene_id) as scene:
            scene.description = req.args.new_description
            scene.update_modified()

            evt = sevts.s.SceneChanged(scene.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))
        return None


async def copy_scene_cb(req: srpc.s.CopyScene.Request, ui: WsClient) -> None:

    async with ctx_write_lock(req.args.source_id, glob.USERS.user_name(ui)):
        # TODO check if target_name is unique
        async with managed_scene(req.args.source_id, make_copy=True) as scene:
            scene.name = req.args.target_name

            evt = sevts.s.SceneChanged(scene.bare)
            evt.change_type = Event.Type.UPDATE_BASE
            asyncio.ensure_future(notif.broadcast_event(evt))

        return None


# TODO maybe this would better fit into another category of RPCs? Like common/misc?
async def get_camera_pose_cb(req: srpc.c.GetCameraPose.Request, ui: WsClient) -> srpc.c.GetCameraPose.Response:

    try:
        return srpc.c.GetCameraPose.Response(
            data=await hlp.run_in_executor(
                calibration.estimate_camera_pose,
                req.args.camera_parameters,
                image_from_str(req.args.image),
                req.args.inverse,
            )
        )
    except calibration.MarkerNotFound:  # this is ok
        raise
    except calibration.CalibrationException as e:  # this means a serious problem and should be logged
        glob.logger.warn(f"Failed to get camera pose. {str(e)}")
        raise


# TODO maybe this would better fit into another category of RPCs? Like common/misc?
async def marker_corners_cb(req: srpc.c.MarkersCorners.Request, ui: WsClient) -> srpc.c.MarkersCorners.Response:

    return srpc.c.MarkersCorners.Response(
        data=await hlp.run_in_executor(
            calibration.markers_corners, req.args.camera_parameters, image_from_str(req.args.image)
        )
    )


async def start_scene_cb(req: srpc.s.StartScene.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()

    if get_scene_state().data.state != sevts.s.SceneState.Data.StateEnum.Stopped:
        raise Arcor2Exception("Scene not stopped.")

    # online scene can't be modified so we demand that UIs free all their locks first
    # when editing project, changes can be done both online and offline
    if not glob.LOCK.project and await glob.LOCK.get_write_locks_count():
        raise LockingException(glob.LOCK.ErrMessages.SOMETHING_LOCKED.value)

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.SCENE_NAME, glob.LOCK.SpecialValues.SERVER_NAME):
        raise Arcor2Exception("Scene locked.")

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.PROJECT_NAME, glob.LOCK.SpecialValues.SERVER_NAME):
        raise Arcor2Exception("Project locked.")

    if req.dry_run:
        return

    asyncio.ensure_future(start_scene(scene))


async def stop_scene_cb(req: srpc.s.StopScene.Request, ui: WsClient) -> None:

    scene = glob.LOCK.scene_or_exception()

    if get_scene_state().data.state != sevts.s.SceneState.Data.StateEnum.Started:
        raise Arcor2Exception("Scene not started.")

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.SCENE_NAME, glob.LOCK.SpecialValues.SERVER_NAME):
        raise Arcor2Exception("Scene locked.")

    if await glob.LOCK.is_write_locked(glob.LOCK.SpecialValues.PROJECT_NAME, glob.LOCK.SpecialValues.SERVER_NAME):
        raise Arcor2Exception("Project locked.")

    if glob.RUNNING_ACTION:  # TODO acquire lock?
        raise Arcor2Exception("There is a running action.")

    if req.dry_run:
        return

    asyncio.ensure_future(stop_scene(scene))
